@@ -379,6 +379,8 @@ class Bounds:
     def union(self, b):
         """ Returns bounds that encompass the union of the two.
         """
+        # Note: this will also work with ghost points.
+        # Bounds(20, 20, 5, 5).union(Bounds(100, 100, 0, 0)) => Bounds(20, 20, 100, 100). 
         mx, my = min(self.x, b.x), min(self.y, b.y)
         return Bounds(mx, my, 
             max(self.x+self.width, b.x+b.width) - mx, 
@@ -447,10 +449,10 @@ class Bounds:
 #  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 from sys import platform
-from ctypes import CFUNCTYPE, POINTER, byref, cast
+from ctypes import CFUNCTYPE, POINTER, byref, cast, pointer
 from ctypes import CFUNCTYPE as _CFUNCTYPE
 from pyglet.gl import \
-    GLdouble, GLvoid, GLenum, GLfloat, pointer, \
+    GLdouble, GLvoid, GLenum, GLfloat, \
     gluNewTess, gluTessProperty, gluTessNormal, gluTessCallback, gluTessVertex, \
     gluTessBeginPolygon, gluTessEndPolygon, \
     gluTessBeginContour, gluTessEndContour, \
