@@ -1,4 +1,4 @@
-#=== SHADER ==========================================================================================
+
 # Fragment shaders, filters, Frame Buffer Object (FBO)
 # Authors: Tom De Smedt, Frederik De Bleser
 # License: BSD (see LICENSE.txt for details).
@@ -10,6 +10,7 @@ from pyglet.image import Texture, SolidColorImagePattern
 from context      import Image, texture
 from geometry     import lerp, clamp
 from math         import radians
+from ctypes       import byref, cast, c_char, c_char_p, c_uint, c_int, pointer, POINTER
 
 def next(generator, default=None):
     try: 
@@ -267,7 +268,6 @@ def shader(vertex=DEFAULT_VERTEX_SHADER, fragment=DEFAULT_FRAGMENT_SHADER, silen
     try:
         return Shader(vertex, fragment)
     except Exception, e:
-        print e
         SUPPORTED = False
         return ShaderFacade()
 
