@@ -58,7 +58,7 @@ class Theme(dict):
         self["fontweight"] = kwargs.get("fontweight", NORMAL)
         self["text"]       = kwargs.get("text", Color(1.0))
 
-theme = Theme(os.path.join(os.path.dirname(__file__), "theme")) 
+theme = Theme(os.path.join(os.path.dirname(os.path.abspath(__file__)), "theme")) 
 
 #=====================================================================================================
 
@@ -66,7 +66,7 @@ theme = Theme(os.path.join(os.path.dirname(__file__), "theme"))
 
 class Control(Layer):
     
-    def __init__(self, x=0, y=0, id=None, **kwargs):
+    def __init__(self, x=0, y=0, id=None, color=(1,1,1,1), **kwargs):
         """ Base class for GUI controls.
             The Control class inherits from Layer so it must be appended to the canvas (or a container)
             to receive events and get drawn.
@@ -76,6 +76,7 @@ class Control(Layer):
         Layer.__init__(self, x=x, y=y, **kwargs)
         self.id        = id
         self.src       = {}    # Collection of source images.
+        self.color     = color # Color for source images.
         self.enabled   = True  # Enable event listener.
         self.duration  = 0     # Disable tweening.
         self._controls = {}    # Lazy index of (id, control) children, see nested().
@@ -237,7 +238,7 @@ class Label(Control):
     def _pack(self):
         # Center the text inside the label.
         self._text.x = 0.5 * (self.width - self._text.metrics[0])
-        self._text.y = 0.5 * (self.height - self._text.metrics[1])  
+        self._text.y = 0.5 * (self.height - self._text.metrics[1])
 
     def draw(self):
         self._text.draw()
@@ -288,7 +289,7 @@ class Button(Control):
         self[0].y = 0.5 * (self.height - self[0].height) - self.pressed
 
     def draw(self):
-        clr = self.pressed and (0.75, 0.75, 0.75) or (1.0, 1.0, 1.0)
+        clr = self.pressed and [v*0.75 for v in self.color] or self.color
         im1, im2, im3 = self.src["cap1"], self.src["cap2"], self.src["face"]
         image(im1, 0, 0, height=self.height, color=clr)
         image(im2, x=self.width-im2.width, height=self.height, color=clr)
@@ -322,7 +323,7 @@ class Action(Control):
         self._set_height(self.src["face"].height)
     
     def draw(self):
-        clr = self.pressed and (0.75, 0.75, 0.75) or (1.0, 1.0, 1.0)
+        clr = self.pressed and [v*0.75 for v in self.color] or self.color
         image(self.src["face"], 0, 0, color=clr)
             
     def on_mouse_release(self, mouse):
@@ -364,7 +365,7 @@ class Handle(Control):
         self.parent.on_mouse_release(mouse)
         
     def draw(self):
-        clr = self.parent.pressed | self.pressed and (0.75, 0.75, 0.75) or (1.0, 1.0, 1.0)
+        clr = self.parent.pressed | self.pressed and [v*0.75 for v in self.color] or self.color
         image(self.parent.src["handle"], 0, 0, color=clr)
 
 class Slider(Control):
@@ -421,10 +422,11 @@ class Slider(Control):
     def draw(self):
         t = self._t * self.width
         im1, im2, im3, im4  = self.src["cap1"], self.src["cap2"], self.src["face1"], self.src["face2"]
-        image(im1, x=0, y=0)
-        image(im2, x=self.width-im2.width, y=0)
-        image(im3, x=im1.width, y=0, width=t-im1.width)
-        image(im4, x=t, y=0, width=self.width-t-im2.width+1)
+        clr = self.color
+        image(im1, x=0, y=0, color=clr)
+        image(im2, x=self.width-im2.width, y=0, color=clr)
+        image(im3, x=im1.width, y=0, width=t-im1.width, color=clr)
+        image(im4, x=t, y=0, width=self.width-t-im2.width+1, color=clr)
 
     def on_mouse_press(self, mouse):
         x0, y0 = self.absolute_position() # Can be nested in other layers.
@@ -476,11 +478,12 @@ class Knob(Control):
         self.value = self.default
 
     def draw(self):
+        clr1 = self.color
+        clr2 = self.pressed and [v*0.85 for v in self.color] or self.color
         translate(self.width/2, self.height/2)
-        image(self.src["socket"], -self.width/2, -self.height/2)
+        image(self.src["socket"], -self.width/2, -self.height/2, color=clr1)
         rotate(360-self.value)
-        clr = self.pressed and (0.85, 0.85, 0.85) or (1.0, 1.0, 1.0)
-        image(self.src["face"], -self.width/2, -self.height/2, color=clr)
+        image(self.src["face"], -self.width/2, -self.height/2, color=clr2)
         
     def on_mouse_press(self, mouse):
         self.value += mouse.dy * (CTRL in mouse.modifiers and 1 or 5)
@@ -518,7 +521,8 @@ class Flag(Control):
         self.value = self.default
     
     def draw(self):
-        image(self.value and self.src["checked"] or self.src["face"])
+        clr = self.color
+        image(self.value and self.src["checked"] or self.src["face"], color=clr)
         
     def on_mouse_release(self, mouse):
         Control.on_mouse_release(self, mouse)
@@ -538,7 +542,7 @@ editing = lambda: EDITING
 
 class Editable(Control):
     
-    def __init__(self, value="", x=0, y=0, width=125, height=30, padding=(0,0), wrap=True, id=None, **kwargs):
+    def __init__(self, value="", x=0, y=0, width=125, height=20, padding=(0,0), wrap=False, id=None, **kwargs):
         """ An editable text box.
             When clicked, it has the focus and can receive keyboard events.
             With wrap=True, several lines of text will wrap around the width.
@@ -549,7 +553,7 @@ class Editable(Control):
                "font" : _popdefault(kwargs, "font", theme["fontname"]),
            "fontsize" : _popdefault(kwargs, "fontsize", theme["fontsize"]),
          "fontweight" : _popdefault(kwargs, "fontweight", theme["fontweight"]),
-         "lineheight" : _popdefault(kwargs, "lineheight", 1),
+         "lineheight" : _popdefault(kwargs, "lineheight", wrap and 1.25 or 1.0),
               "align" : LEFT
         })
         kwargs["width"]  = width
@@ -686,8 +690,8 @@ class Editable(Control):
                 y = keys.code == UP and -1 or +1
                 n = self._editor.get_line_count()
                 i = self._editor.get_position_on_line(
-                    max(self._editor.get_line_from_position(i)+y, 0),
-                        self._editor.get_point_from_position(i)[0])
+                    min(max(self._editor.get_line_from_position(i)+y, 0), n-1),
+                            self._editor.get_point_from_position(i)[0])
                 self._editor.caret.position = i
             elif keys.code == TAB and TAB in self.reserved:
                 # The tab key navigates away from the control.
@@ -728,12 +732,18 @@ class Field(Editable):
         """ A single-line text input field.
             The string value can be retrieved with Field.value.
         """
-        Editable.__init__(self, value, x=x, y=y, width=width, padding=(padding,0), wrap=False, id=id, **kwargs)
+        Editable.__init__(self, value, x=x, y=y, width=width, padding=[padding]*2, id=id, **kwargs)
         img, w = Image(theme["field"]), 10
         self.src = {
-            "face" : crop(img, w, 0, 1, img.height),
-            "cap1" : crop(img, 0, 0, w, img.height),
-            "cap2" : crop(img, img.width-w, 0, w, img.height),
+           "cap1" : crop(img, 0, img.height-w, w, w),
+           "cap2" : crop(img, img.width-w, img.height-w, w, w),
+           "cap3" : crop(img, 0, 0, w, w),
+           "cap4" : crop(img, img.width-w, 0, w, w),
+            "top" : crop(img, w+1, img.height-w, 1, w),
+         "bottom" : crop(img, w+1, 0, 1, w),
+           "left" : crop(img, 0, w+1, w, 1),
+          "right" : crop(img, img.width-w, w+1, w, 1),
+           "face" : crop(img, w+1, w+1, 1, 1)
         }
         if action:
             # Override the Button.on_action() method from the given function.
@@ -755,11 +765,15 @@ class Field(Editable):
     def _pack(self):
         Editable._pack(self)
         w = max(self.width, self.src["cap1"].width + self.src["cap2"].width)
+        h = max(self.height, self.src["cap1"].width + self.src["cap3"].width)
+        h = max(h, int(self._editor.document.get_style("line_spacing") * 1.5 + self._padding[1] * 2))
         self._set_width(w)
-        self._set_height(self.src["face"].height)
-        # Position the hint text (if no other text is in the field):
+        self._set_height(h)
+        # Position the hint text (if no other text is in the field).
+        # The hint will not span multiple line if it is wider than the field
+        # (it was designed to be a short word or phrase).
         self[0].x = self._padding[0]
-        self[0]._set_height(self.height)
+        self[0].y = self.height - self._padding[1] - self[0]._text.metrics[1] * 1.25
         self[0]._pack()
     
     def on_action(self):
@@ -769,10 +783,19 @@ class Field(Editable):
         self[0].hidden = self.editing or self.value != ""
     
     def draw(self):
-        im1, im2, im3 = self.src["cap1"], self.src["cap2"], self.src["face"]
-        image(im1, 0, 0, height=self.height)
-        image(im2, x=self.width-im2.width, height=self.height)
-        image(im3, x=im1.width, width=self.width-im1.width-im2.width, height=self.height)
+        im1, im2, im3 = self.src["cap1"], self.src["cap2"],  self.src["top"]
+        im4, im5, im6 = self.src["cap3"], self.src["cap4"],  self.src["bottom"]
+        im7, im8, im9 = self.src["left"], self.src["right"], self.src["face"]
+        clr = self.color
+        image(im1, 0, self.height-im1.height, color=clr)
+        image(im2, self.width-im2.width, self.height-im2.height, color=clr)
+        image(im3, im1.width, self.height-im3.height, width=self.width-im1.width-im2.width, color=clr)
+        image(im4, 0, 0, color=clr)
+        image(im5, self.width-im5.width, 0, color=clr)
+        image(im6, im4.width, 0, width=self.width-im4.width-im5.width, color=clr)
+        image(im7, 0, im4.height, height=self.height-im1.height-im4.height, color=clr)
+        image(im8, self.width-im8.width, im4.height, height=self.height-im2.height-im5.height, color=clr)
+        image(im9, im4.width, im6.height, width=self.width-im7.width-im8.width, height=self.height-im3.height-im6.height, color=clr)
         Editable.draw(self)
 
 #=====================================================================================================
@@ -781,7 +804,7 @@ class Field(Editable):
 
 class Rulers(Control):
     
-    def __init__(self, step=10, interval=5, crosshair=False, fill=(0,0,0,1)):
+    def __init__(self, step=10, interval=5, crosshair=False, color=(0,0,0,1)):
         """ A horizontal and vertical ruler displaying the width/height of the parent at intervals.
             A measurement line is drawn at each step(e.g. at 10 20 30...)
             A label with the value is drawn at each interval (e.g. 50 | | | | 100 | | | | 150).
@@ -791,7 +814,7 @@ class Rulers(Control):
         self.step      = step
         self.interval  = interval
         self.crosshair = crosshair
-        self._fill     = fill
+        self.color     = color
         self._dirty    = False
         self._markers  = {}
         self._pack()
@@ -827,7 +850,7 @@ class Rulers(Control):
                         Text(str(int(round(i*self._step))), 
                             fontname = theme["fontname"],
                             fontsize = theme["fontsize"] * 0.6,
-                                fill = self._fill))
+                                fill = self.color))
                             
     def update(self):
         self._pack()
@@ -838,7 +861,7 @@ class Rulers(Control):
         for i in range(1, int(round(self.height / self._step))):
             v, mark = i*self._step, i%self.interval==0
             line(0, v, mark and length*3 or length, v, 
-                     stroke = self._fill, 
+                     stroke = self.color, 
                 strokewidth = 0.5)
             if mark:
                 self._markers[v].draw(length*3-self._markers[v].metrics[0], v+2)
@@ -846,18 +869,18 @@ class Rulers(Control):
         for i in range(1, int(round(self.width / self._step))):
             v, mark = i*self._step, i%self.interval==0
             line(v, 0, v, mark and length*3 or length, 
-                     stroke = self._fill, 
+                     stroke = self.color, 
                 strokewidth = 0.5)
             if mark:
                 self._markers[v].draw(v+2, length*3-self._markers[v].fontsize)
         # Draw the crosshair.
         if self.crosshair:
             line(0, self.canvas.mouse.y, self.width, self.canvas.mouse.y, 
-                     stroke = self._fill, 
+                     stroke = self.color, 
                 strokewidth = 0.5, 
                 strokestyle = DOTTED)
             line(self.canvas.mouse.x, 0, self.canvas.mouse.x, self.height, 
-                     stroke = self._fill, 
+                     stroke = self.color, 
                 strokewidth = 0.5, 
                 strokestyle = DOTTED)
 
@@ -886,8 +909,9 @@ class Panel(Control):
         }
         _popdefault(kwargs, "width")
         _popdefault(kwargs, "height")
-        self.append(Label(caption, **kwargs))
+        self.append(Label(caption))
         self.append(Close())
+        #self.extend(kwargs.pop("controls", []), **kwargs)
         self.fixed = fixed # Draggable?
         self.modal = modal # Closeable?
         self._pack()
@@ -945,21 +969,23 @@ class Panel(Control):
     
     def update(self):
         self[1].hidden = self.modal
+        self[1].color  = self.color
     
     def draw(self):
         im1, im2, im3 = self.src["cap1"], self.src["cap2"],  self.src["top"]
         im4, im5, im6 = self.src["cap3"], self.src["cap4"],  self.src["bottom"]
         im7, im8, im9 = self.src["left"], self.src["right"], self.src["face"]
-        image(im1, 0, self.height-im1.height)
-        image(im2, self.width-im2.width, self.height-im2.height)
-        image(im3, im1.width, self.height-im3.height, width=self.width-im1.width-im2.width)
-        image(im4, 0, 0)
-        image(im5, self.width-im5.width, 0)
-        image(im6, im4.width, 0, width=self.width-im4.width-im5.width)
-        image(im7, 0, im4.height, height=self.height-im1.height-im4.height)
-        image(im8, self.width-im8.width, im4.height, height=self.height-im2.height-im5.height)
-        image(im9, im4.width, im6.height, width=self.width-im7.width-im8.width, height=self.height-im3.height-im6.height)
-            
+        clr = self.color
+        image(im1, 0, self.height-im1.height, color=clr)
+        image(im2, self.width-im2.width, self.height-im2.height, color=clr)
+        image(im3, im1.width, self.height-im3.height, width=self.width-im1.width-im2.width, color=clr)
+        image(im4, 0, 0, color=clr)
+        image(im5, self.width-im5.width, 0, color=clr)
+        image(im6, im4.width, 0, width=self.width-im4.width-im5.width, color=clr)
+        image(im7, 0, im4.height, height=self.height-im1.height-im4.height, color=clr)
+        image(im8, self.width-im8.width, im4.height, height=self.height-im2.height-im5.height, color=clr)
+        image(im9, im4.width, im6.height, width=self.width-im7.width-im8.width, height=self.height-im3.height-im6.height, color=clr)
+        
     def on_mouse_enter(self, mouse): 
         mouse.cursor = DEFAULT
 
@@ -979,7 +1005,7 @@ class Panel(Control):
 
 class Dock(Panel):
     
-    def __init__(self, anchor=LEFT, caption="", fixed=True, modal=True, **kwargs):
+    def __init__(self, caption="", anchor=LEFT, fixed=True, modal=True, **kwargs):
         """ A panel attached to the edge of the canvas (LEFT or RIGHT), extending the full height.
             With fixed=False, it can be snapped from the edge and dragged as a normal panel.
         """
@@ -1009,8 +1035,9 @@ class Dock(Panel):
         if self.canvas is not None and \
           (self.anchor == LEFT  and self.x == 0) or \
           (self.anchor == RIGHT and self.x == self.canvas.width-self.width):
-            image(im1, 0, self.height-im1.height, width=self.width)
-            image(im2, 0, -self.canvas.height+self.height, width=self.width, height=self.canvas.height-im1.height)
+            clr = self.color
+            image(im1, 0, self.height-im1.height, width=self.width, color=clr)
+            image(im2, 0, -self.canvas.height+self.height, width=self.width, height=self.canvas.height-im1.height, color=clr)
         else:
             Panel.draw(self)
 
@@ -1020,21 +1047,27 @@ class Dock(Panel):
 
 class Layout(Layer):
     
-    def __init__(self, x=0, y=0, **kwargs):
+    SPACING = 10 # Spacing between controls in a Layout.
+    
+    def __init__(self, controls=[], x=0, y=0, **kwargs):
         """ A group of controls with a specific layout.
             Controls can be added with Layout.append().
             The layout will be applied when Layout.apply() is called.
             This happens automatically if a layout is appended to a Panel.
         """
-        
         kwargs["width"]  = 0
         kwargs["height"] = 0
         Layer.__init__(self, x=x, y=y, **kwargs)
         self._controls = {} # Lazy cache of (id, control)-children, see nested().
+        self.spacing = kwargs.get("spacing", Layout.SPACING)
+        self.extend(controls)
 
     def insert(self, i, control):
+        """ Inserts the control, or inserts all controls in the given Layout.
+        """
         if isinstance(control, Layout):
-            control.apply() # If the control is actually a Layout, apply it.
+            # If the control is actually a Layout (e.g. ordered group of controls), apply it.
+            control.apply()
         Layer.insert(self, i, control)
         
     def append(self, control):
@@ -1060,26 +1093,27 @@ class Layout(Layer):
             return ctrl
         raise AttributeError, "'%s' object has no attribute '%s'" % (self.__class__.__name__, k)
 
-    def apply(self, spacing=0):
+    def apply(self):
         """ Adjusts the position and size of the controls to match the layout.
         """
-        pass
+        self.width  = max(control.width  for control in self)
+        self.height = max(control.height for control in self)
         
     def __repr__(self):
         return "Layout(type=%s)" % repr(self.__class__.__name__.lower())
     
     # Debug mode:
     #def draw(self):
-    #    rect(0, 0, self.width, self.height, fill=None, stroke=(1,1,1,0.5), strokestyle="dotted")
+    #    rect(0, 0, self.width, self.height, fill=None, stroke=(0,0.5,1,1))
 
 #--- Layout: Labeled ----------------------------------------------------------------------------------
 
 class Labeled(Layout):
 
-    def __init__(self, controls=[], x=0, y=0):
+    def __init__(self, controls=[], x=0, y=0, **kwargs):
         """ A layout where each control has an associated text label.
         """
-        Layout.__init__(self, x=x, y=y)
+        Layout.__init__(self, controls=[], x=x, y=y, **kwargs)
         self.controls = []
         self.captions = []
         self.extend(controls)
@@ -1111,15 +1145,15 @@ class Labeled(Layout):
 
 class Rows(Labeled):
 
-    def __init__(self, controls=[], x=0, y=0, width=125):
+    def __init__(self, controls=[], x=0, y=0, width=125, **kwargs):
         """ A layout where each control appears on a new line.
             Each control has an associated text caption, displayed to the left of the control.
             The given width defines the desired width for each control.
         """
-        Labeled.__init__(self, controls, x=x, y=y)
+        Labeled.__init__(self, controls, x=x, y=y, **kwargs)
         self._maxwidth = width
 
-    def apply(self, spacing=10):
+    def apply(self):
         """ Adjusts the position and width of all the controls in the layout:
             - each control is placed next to its caption, with spacing in between,
             - each caption is aligned to the right, and centered vertically,
@@ -1129,59 +1163,60 @@ class Rows(Labeled):
         for control in self.controls:
             if isinstance(control, Layout):
                 # Child containers in the layout can be wider than the desired width.
-                # adjusting mw at the start will controls wider to line out with the total width,
+                # adjusting mw at the start will make controls wider to line out with the total width,
                 # adjusting it at the end would just ensure that the layout is wide enough.
-                mw = max(self._maxwidth, control.width)
-        w1 = max([caption.width for caption in self.captions])
-        w2 = max([control.width for control in self.controls])
+                mw = max(mw, control.width)
+        w1 = max(caption.width for caption in self.captions)
+        w2 = max(control.width for control in self.controls)
+        w2 = min(w2, mw)
         dx = 0
         dy = 0
         for caption, control in reversed(zip(self.captions, self.controls)):
-            caption.x = dx + w1 - caption.width                      # halign right.
-            control.x = dx + w1 + (w1>0 and spacing)
-            caption.y = dy + 0.5 * (control.height - caption.height) # valign center.
-            control.y = dy
             if isinstance(control, Layout) and control.height > caption.height * 2:
                 caption.y = dy + control.height - caption.height     # valign top.
             if isinstance(control, (Label, Button, Slider, Field)):
                 control._set_width(mw)
                 control._pack()
-            dy += max(caption.height, control.height, 10) + spacing
-        self.width  = w1 + w2 + (w1>0 and spacing)
-        self.height = dy - spacing
-
-TOP, CENTER = "top", "center"
+            caption.x = dx + w1 - caption.width                      # halign right.
+            control.x = dx + w1 + (w1>0 and self.spacing)
+            caption.y = dy + 0.5 * (control.height - caption.height) # valign center.
+            control.y = dy
+            dy += max(caption.height, control.height) + self.spacing
+        self.width  = w1 + max(w2, mw) + (w1>0 and self.spacing)
+        self.height = dy - self.spacing
+        
+TOP, BOTTOM, CENTER = "top", "bottom", "center"
 
 class Row(Labeled):
 
-    def __init__(self, controls=[], x=0, y=0, width=125, align=CENTER):
+    def __init__(self, controls=[], x=0, y=0, width=125, align=CENTER, **kwargs):
         """ A layout where each control appears in a new column.
             Each control has an associated text caption, displayed on top of the control.
             The given width defines the desired width for each control.
         """
-        Labeled.__init__(self, controls, x=x, y=y)
+        Labeled.__init__(self, controls, x=x, y=y, **kwargs)
         self._maxwidth = width
         self._align    = align
 
-    def apply(self, spacing=10):
+    def apply(self):
         """ Adjusts the position and width of all the controls in the layout:
             - each control is placed centrally below its caption, with spacing in between,
             - the width of all Label, Button, Slider, Field controls is evened out.
         """
         mw = self._maxwidth
-        da = self._align==TOP and 1.0 or 0.5
-        h1 = max([control.height for control in self.controls])
-        h2 = max([caption.height for caption in self.captions])
+        da = {TOP: 1.0, BOTTOM: 0.0, CENTER: 0.5}.get(self._align, 0.5)
+        h1 = max(control.height for control in self.controls)
+        h2 = max(caption.height for caption in self.captions)
         dx = 0
         dy = 0
         for caption, control in zip(self.captions, self.controls):
-            caption.x = dx + 0.5 * max(control.width - caption.width, 0) # halign center
-            control.x = dx + 0.5 * max(caption.width - control.width, 0) # halign center
-            caption.y = dy + h1 + (h2>0 and spacing)                 
-            control.y = dy + da * (h1 - control.height)                  # valign center
             if isinstance(control, (Label, Button, Slider, Field)):
                 control._set_width(mw)
                 control._pack()
-            dx += max(caption.width, control.width, 10) + spacing
-        self.width = dx - spacing
-        self.height = h1 + h2 + (h2>0 and spacing)
+            caption.x = dx + 0.5 * max(control.width - caption.width, 0) # halign center
+            control.x = dx + 0.5 * max(caption.width - control.width, 0) # halign center
+            caption.y = dy + h1 + (h2>0 and self.spacing)                 
+            control.y = dy + da * (h1 - control.height)                  # valign center
+            dx += max(caption.width, control.width) + self.spacing
+        self.width = dx - self.spacing
+        self.height = h1 + h2 + (h2>0 and self.spacing)
